/* 
  22.03.2021 Antonio Velkov
  Modificado por:
  SETA43 1/10/2024
  Para que funcione con la placa LOLIN32 LITE  y SSD1306 con I2C
  seta43.duckdns.org
  seta43.blogspot.com
*/


// Funcionamiento "contickerDecisec.attach_ms" o con "esp_timer_get_time()"
#define MODO_INTERRUPCION

#include <WiFi.h>
#include <Ticker.h>

#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>

#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels

// Declaration for an SSD1306 display connected to I2C (SDA, SCL pins)
// Wire.begin(23, 19);  //Para esp32 lolin32 mini --Wire.begin(sda_pin, scl_pin)
#define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)
#define SCREEN_ADDRESS 0x3C ///< See datasheet for Address; 0x3D for 128x64, 0x3C for 128x32
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);


// Wifi credentials
const char* ssid = "yourSSID";
const char* password = "yourPassword";

int impulseArray[60];
int impulseCount = 0;
int actualHours, actualMinutes, actualSecond, actualDay, actualMonth, actualYear, DayOfW;
int timeRunningContinuous = 0;

const char* ntpServer = "bg.pool.ntp.org"; // enter your closest pool or pool.ntp.org
const char* TZ_INFO    = "CET-1CEST-2,M3.5.0/02:00:00,M10.5.0/03:00:00";  // enter your time zone (https://remotemonitoringsystems.ca/time-zone-abbreviations.php)

struct tm timeinfo;
String time_data;

Ticker tickerDecisec; // TBD at 100ms

#define LEDBUILTIN 26
#define ANTENNAPIN 25
#define FRECUENCIA 77500
#define RESOLUCION 8


char VARI[80];

void setup() 
{
  Serial.begin(9600);
  Serial.println("");
  Serial.println("INICIO DCF77 transmitter A. Velkov");
  Wire.begin(23, 19);  //Para esp32 lolin32 mini --Wire.begin(sda_pin, scl_pin)

  // SSD1306_SWITCHCAPVCC = generate display voltage from 3.3V internally
  if(!display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS)) {
    Serial.println(F("SSD1306 allocation failed"));
    for(;;); // Don't proceed, loop forever
  }


  display.clearDisplay();
  display.setTextSize(1); // Draw 2X-scale text
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 0);
  display.println(F("DCF77 transmitter"));
  display.println(F("A. Velkov"));
  display.display();      // Show initial text

  ledcAttach(ANTENNAPIN, FRECUENCIA, RESOLUCION);
  delay(1);
  ledcWrite(ANTENNAPIN, 128); //configuramos el Duty Cycle al 50% del led LEDPWM1
  pinMode (LEDBUILTIN, OUTPUT); 


  WiFi_on();
  getNTP();
  WiFi_off();
  CodeTime();

  
  Serial.print(&timeinfo, "Time now: %B %d %Y %H:%M:%S (%A) %Z ");
  if (timeinfo.tm_isdst == 0) 
        { Serial.println("DST=OFF"); } 
    else
        { Serial.println("DST=ON");}

#ifdef MODO_INTERRUPCION  
  tickerDecisec.attach_ms(100, DcfOut); // from now on calls DcfOut every 100ms
#endif  
}

int64_t tiempo,tiempoOld;

void loop() 
{

#ifndef MODO_INTERRUPCION
  //tiempo=millis(); no funciona bien ESP32
  tiempo=esp_timer_get_time();
  if(tiempo-tiempoOld>100000 )
    {
      tiempoOld=tiempo;
      DcfOut();
    }
#endif
}



void CodeTime() 
{
  DayOfW = timeinfo.tm_wday;
  if (DayOfW == 0) DayOfW = 7;
  actualDay = timeinfo.tm_mday;
  actualMonth = timeinfo.tm_mon + 1;
  actualYear = timeinfo.tm_year - 100;
  actualHours = timeinfo.tm_hour;
  actualMinutes = timeinfo.tm_min + 1; // DCF77 transmitts the next minute
  if (actualMinutes >= 60) {
    actualMinutes = 0;
    actualHours++;
  }
  actualSecond = timeinfo.tm_sec; 
  if (actualSecond == 60) actualSecond = 0;

  sprintf(VARI,"%02d/%02d/%02d  %02d:%02d",actualDay,actualMonth,actualYear,actualHours,actualMinutes); 
  display.clearDisplay();
  display.setCursor(0,0);
  display.print("Proximo minuto  ");
  display.println(60-actualSecond);
  display.setCursor(0,20);
  display.println(VARI);
  display.display();

  int n, Tmp, TmpIn;
  int ParityCount = 0;

  //we put the first 20 bits of each minute at a logical zero value
  for (n = 0; n < 20; n++) impulseArray[n] = 1;
  
  // set DST bit
  if (timeinfo.tm_isdst == 0) {
    impulseArray[18] = 2; // CET or DST OFF
  } else {
    impulseArray[17] = 2; // CEST or DST ON
  }
  
  //bit 20 must be 1 to indicate active time
  impulseArray[20] = 2;

  //calculates the bits for the minutes
  TmpIn = Bin2Bcd(actualMinutes);
  for (n = 21; n < 28; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  if ((ParityCount & 1) == 0)
    impulseArray[28] = 1;
  else
    impulseArray[28] = 2;

  //calculates bits for the hours
  ParityCount = 0;
  TmpIn = Bin2Bcd(actualHours);
  for (n = 29; n < 35; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  if ((ParityCount & 1) == 0)
    impulseArray[35] = 1;
  else
    impulseArray[35] = 2;
  ParityCount = 0;

  //calculate the bits for the actual Day of Month
  TmpIn = Bin2Bcd(actualDay);
  for (n = 36; n < 42; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  TmpIn = Bin2Bcd(DayOfW);
  for (n = 42; n < 45; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  //calculates the bits for the actualMonth
  TmpIn = Bin2Bcd(actualMonth);
  for (n = 45; n < 50; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  //calculates the bits for actual year
  TmpIn = Bin2Bcd(actualYear);   // 2 digit year
  for (n = 50; n < 58; n++) {
    Tmp = TmpIn & 1;
    impulseArray[n] = Tmp + 1;
    ParityCount += Tmp;
    TmpIn >>= 1;
  }
  //equal date
  if ((ParityCount & 1) == 0)
    impulseArray[58] = 1;
  else
    impulseArray[58] = 2;

  //last missing pulse
  impulseArray[59] = 0; // No pulse
}

int Bin2Bcd(int dato) {
  int msb, lsb;
  if (dato < 10)
    return dato;
  msb = (dato / 10) << 4;
  lsb = dato % 10;
  return msb + lsb;
}

void DcfOut() {
  switch (impulseCount++) {
    case 0:
      if (impulseArray[actualSecond] != 0) {
        digitalWrite(LEDBUILTIN, LOW);
        ledcWrite(0, 0);
      }
      break;
    case 1:
      if (impulseArray[actualSecond] == 1) {
        digitalWrite(LEDBUILTIN, HIGH);
        ledcWrite(0, 127);
      }
      break;
    case 2:
      digitalWrite(LEDBUILTIN, HIGH);
      ledcWrite(0, 127);
      break;
    case 9:
      impulseCount = 0;

      if (actualSecond == 1 || actualSecond == 15 || actualSecond == 21  || actualSecond == 29 ) {
        Serial.print("-");
        time_data+="-";
      }
      if (actualSecond == 36  || actualSecond == 42 || actualSecond == 45  || actualSecond == 50 ) {
        Serial.print("-");
        time_data+="-";
      }
      if (actualSecond == 28  || actualSecond == 35  || actualSecond == 58 ) {
        Serial.print("P");
        time_data+="P";
      }

      if (impulseArray[actualSecond] == 1) {
        Serial.print("0");
        time_data+="0";
      }
      if (impulseArray[actualSecond] == 2) {
        Serial.print("1");
        time_data+="1";
      }

      if (actualSecond == 59 ) {
        Serial.println();
                
        time_data="";

        timeRunningContinuous++;
        if (timeRunningContinuous > 360) ESP.restart(); // 6 hours running, then restart all over

      }
      break;
  }
  if(!getLocalTime(&timeinfo)){
    Serial.println("Error obtaining time...");
    display.clearDisplay();
    display.setCursor(0, 0);
    display.println(F("Error obtaining time..."));
    display.display();
    delay(3000);
    ESP.restart();
  }
  CodeTime();
}



void getNTP() 
{
  Serial.print("GetNTP ");
  display.clearDisplay();
  display.setCursor(0, 0);
  display.println(F("Syncing NTP..."));
  int i = 0;
  do {
    i++;
    if (i > 40) {
      ESP.restart();
    }
    configTime(0, 0, ntpServer);
    setenv("TZ", TZ_INFO, 1);
    delay(500);
  } while (!getLocalTime(&timeinfo));
  Serial.println("Ok");
  display.println();
  display.println(F("Done!"));
  display.display();
  delay(2000);
}


void WiFi_on() 
{
  
  byte mac[6];
  
  Serial.print("Connecting to: "+String(ssid));
  display.clearDisplay();
  display.setCursor(0, 0);
  display.println(F("Connecting WiFi to:"));
  display.println(ssid);
  display.println("...");
  display.display();
  delay(500);
  
  WiFi.begin(ssid, password);
  int counter = 0;
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    if (counter > 20) ESP.restart();
    Serial.print ( "." );
    counter++;
  }
  Serial.println();
  Serial.println("WiFi connected");
  display.clearDisplay();
  display.setCursor(0, 0);
  display.println(F("Connecting WiFi to:"));
  display.println(ssid);
  display.println(F("Connected!"));
  display.println();
  display.println(WiFi.localIP());
  display.println();
  WiFi.macAddress(mac);
  display.print(mac[5],HEX);
  display.print(":");
  display.print(mac[4],HEX);
  display.print(":");
  display.print(mac[3],HEX);
  display.print(":");
  display.print(mac[2],HEX);
  display.print(":");
  display.print(mac[1],HEX);
  display.print(":");
  display.println(mac[0],HEX);
  
  display.display();
  delay(5000);
}

void WiFi_off() {
  WiFi.disconnect(true);
  WiFi.mode(WIFI_OFF);
  Serial.println("WiFi disconnected");
  Serial.flush();
}
